#pragma pack(4)
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#include <stdio.h>
#include <conio.h>

#include <stdlib.h>
#include <winsock.h>

#include "linkedlist.h"


void WriteLog(unsigned char * Packet, int Length)
{
	printf("S2C 0000  ");

	for(int cnt = 0; cnt < Length; cnt++)
	{
		if(cnt && cnt % 16 == 0)
			printf("\nS2C %04X  ", cnt);
		printf("%02X ", Packet[cnt]);
	}
	printf("\n\n");
}


void ReadLog(unsigned char * Packet, int Length)
{
	printf("C2S 0000  ");

	for(int cnt = 0; cnt < Length; cnt++)
	{
		if(cnt && cnt % 16 == 0)
			printf("\nC2S %04X  ", cnt);
		printf("%02X ", Packet[cnt]);
	}
	printf("\n\n");
}


class Mutex
{
	HANDLE hMutex;

public:
	Mutex()
	{
		hMutex = CreateMutex(NULL, false, NULL);
	}

	virtual ~Mutex()
	{
		CloseHandle(hMutex);
		hMutex = NULL;
	}

	void LockClass() const
	{
		WaitForSingleObject(hMutex, INFINITE);
	}

	void UnlockClass() const
	{
		ReleaseMutex(hMutex);
	}

	inline Mutex(const Mutex &copy)
	{
		hMutex = copy.hMutex;
	}
};


class Host
{
	friend class WinsockSession;

	sockaddr_in Remote;
	SOCKET SocketID;
	HANDLE hProxy;
	DWORD LastTicks;
	WinsockSession * ws;

	void ProxyThread();	// Receive messages from server-side

	friend DWORD WINAPI CallProxyThread(Host * ThisInstance)
	{
		ThisInstance->ProxyThread();

		return 0;
	}

public:
	bool ValidateHost(sockaddr_in ClientStack)
	{
		if ((ClientStack.sin_addr.S_un.S_addr == Remote.sin_addr.S_un.S_addr) && (ClientStack.sin_port == Remote.sin_port))
			return true;
		else
			return false;
	}

	void Send(BYTE * buffer, DWORD datalen);

	Host(sockaddr_in * ClientStack, WinsockSession * HostStack)
	{
		Remote = *ClientStack;
		ws = HostStack;

		sockaddr_in sa;
					sa.sin_addr.S_un.S_addr	= 0;
					sa.sin_family			= 2;
					sa.sin_port				= 0;

		SocketID = socket(2, 2, 0);
		bind(SocketID, (sockaddr *) &sa, 16);

		DWORD iThread;
		hProxy = CreateThread(NULL, 0, (DWORD (__stdcall *)(void *))&CallProxyThread, this, 0, &iThread);
	}

	~Host()
	{
		DWORD ExitCode;

		GetExitCodeThread(hProxy, &ExitCode);
		TerminateThread(hProxy, ExitCode);

		CloseHandle(hProxy);

		closesocket(SocketID);
	}
};


class WinsockSession
{
	HANDLE hReceive;

	_linkedlist<Host> HostList;

	void RecvThread()	// Receive messages from client-side
	{
		sockaddr_in SockAddy;
		int FromLen = 16;
		BYTE buffer[520];

		while (hReceive)
		{
			int datalen = recvfrom(ServerSocket, (char*)buffer, 520, 0, (sockaddr*)&SockAddy, &FromLen);

			if (datalen > 0)
			{
				// Look for source client
				_listnode<Host> * parse = HostList.head;
				Host * host;

				while (parse)
				{
					host = parse->item;

					if (host->ValidateHost(SockAddy))
					{
						host->Send(buffer, datalen);

						break;
					}

					parse = parse->next;
				}

				// Create a new client
				if (parse == NULL)
				{
					host = new Host(&SockAddy, this);

					host->Send(buffer, datalen);

					HostList.Prepend(host);
				}
			}
		}
	}

	friend DWORD WINAPI CallRecvThread(WinsockSession * ThisInstance)
	{
		ThisInstance->RecvThread();

		return 0;
	}

public:
	SOCKET ServerSocket;
	sockaddr_in ServerAddress;

	/* Initialization */
	void Initialize(SHORT NewPort)
	{
		// Kill old session
		if (hReceive)
			Terminate();

		// Create session with Winsock
		sockaddr_in sa;
					sa.sin_addr.S_un.S_addr	= 0;
					sa.sin_family			= 2;
					sa.sin_port				= htons(NewPort);

		ServerSocket = socket(2, 2, 0);
		bind(ServerSocket, (sockaddr *) &sa, 16);

		// Start threads
		DWORD iThread;

		hReceive = CreateThread(NULL, 0, (DWORD (__stdcall *)(void *))&CallRecvThread, this, 0, &iThread);
	}

	void Terminate()
	{
		// Kill threads
		DWORD ExitCode;

		GetExitCodeThread(hReceive, &ExitCode);
		TerminateThread(hReceive, ExitCode);

		// Unregister threads
		CloseHandle(hReceive);

		// Unregister socket
		closesocket(ServerSocket);

		hReceive = 0;
	}

	WinsockSession()
	{
		hReceive = 0;
	}

	~WinsockSession()
	{
		if (hReceive)
			Terminate();
	}
};


LONG ResolveHostname(char * Hostname)
{
	LONG RetIP = inet_addr(Hostname);

	if (RetIP == -1)
	{
		hostent * host = gethostbyname(Hostname);

		if (host)
			RetIP = *((LONG*)*(host->h_addr_list));
		else
			RetIP = -1;
	}

	return RetIP;
}


void main()
{
	WSADATA wsaData;
	WSAStartup(0x202, &wsaData);

	char buffer[64];

	// Load settings -> Set up server mapping
	char Path[520];
	GetCurrentDirectory(520, Path);

	// Log file
	char Logfile[520];
	memset(buffer, 0, 64);
	GetPrivateProfileString("Proxy", "Log", "packets.txt", buffer, 60, Path);
	strcpy(Logfile, Path);
	strcat(Logfile, "\\");
	strcat(Logfile, buffer);

	strcat(Path, "\\SSPROXY.INI");

	memset(buffer, 0, 32);
	GetPrivateProfileString("Proxy", "Port", "15000", buffer, 16, Path);
	SHORT Port = atoi(buffer);

	// Start winsock
	WinsockSession ws;
	WinsockSession ws_ping;
	ws.Initialize(Port);
	ws_ping.Initialize(Port + 1);

	memset(buffer, 0, 32);
	GetPrivateProfileString("Server", "Port", "2000", buffer, 16, Path);
		ws.ServerAddress.sin_port				= htons(atoi(buffer));
		ws_ping.ServerAddress.sin_port			= htons(atoi(buffer) + 1);
	printf("Server Port = %i\n", atoi(buffer));

	memset(buffer, 0, 64);
	GetPrivateProfileString("Server", "IP", "127.0.0.1", buffer, 60, Path);
		LONG IP = ResolveHostname(buffer);
		ws.ServerAddress.sin_addr.S_un.S_addr	= IP;
		ws_ping.ServerAddress.sin_addr.S_un.S_addr	= IP;
	printf("Server IP = %i.%i.%i.%i\n", (int)((unsigned char*)&IP)[0]
									  , (int)((unsigned char*)&IP)[1]
									  , (int)((unsigned char*)&IP)[2]
									  , (int)((unsigned char*)&IP)[3]);

		ws.ServerAddress.sin_family				= 2;
		ws_ping.ServerAddress.sin_family		= 2;

	printf("SSProxy BUILD 6 - Ready to rock on port %i!\n", Port);
	printf("NOTE: Press any key to terminate.\n");

	freopen(Logfile, "wt", stdout);

	// Close on kbhit
	while (!kbhit())
	{
		Sleep(500);
	};

	// Bye!
	ws.Terminate();
	ws_ping.Terminate();

	WSACleanup();
}

void Host::Send(BYTE * buffer, DWORD datalen)
{
	sendto(SocketID, (char*)&buffer[0], datalen, 0, (sockaddr*)&ws->ServerAddress, 16);

//	ReadLog(buffer, datalen);
}

void Host::ProxyThread()	// Receive messages from server-side
{
	sockaddr_in SockAddy;
	int FromLen = 16;
	BYTE buffer[520];

	while (hProxy)
	{
		int datalen = recvfrom(SocketID, (char*)buffer, 520, 0, (sockaddr*)&SockAddy, &FromLen);

		if (datalen > 0)
		{
			sendto(ws->ServerSocket, (char*)&buffer[0], datalen, 0, (sockaddr*)&Remote, 16);

//			WriteLog(buffer, datalen);
		}
	}
}
